
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/select.h>

#include "select.h"

#define SELECT_SLOTS_COUNT 16

static struct {
	int active;

	int fd;

	select_read_cb  read_cb;
	select_write_cb write_cb;
	select_close_cb close_cb;

	int read_wanted;
	int write_wanted;

	void *ref;

} _select[SELECT_SLOTS_COUNT];

void select_set_read_wanted(const int i, const int yes) {
	_select[i].read_wanted = yes;
}

void select_set_write_wanted(const int i, const int yes) {
	_select[i].write_wanted = yes;
}

int select_register(const int fd, select_read_cb read_cb, select_write_cb write_cb, select_close_cb close_cb, void * const ref) {
	int i;

	for(i = 0; i < SELECT_SLOTS_COUNT; i++) {
		// find an empty slot 
		if(_select[i].active) continue;

		// store fd to monitor
		_select[i].fd       = fd;
		_select[i].read_cb  = read_cb;
		_select[i].write_cb = write_cb;
		_select[i].close_cb = close_cb;
		_select[i].ref      = ref;

		// by default reads are wanted, but writes not
		_select[i].read_wanted  = 1;
		_select[i].write_wanted = 0;

		// mark this slot active
		_select[i].active = 1;

		// all good
		return(i);
	}

	// no free slot found
	return(-1);
}

int select_tick(const int usec_timeout) {
	fd_set rs, ws, es;

	// clear read, write and except set
	FD_ZERO(&rs);
	FD_ZERO(&ws);
	FD_ZERO(&es);

	// store all fds to monitor in sets
	int i, max = 0;
	for(i = 0; i < SELECT_SLOTS_COUNT; i++) {
		// skip inactive slots
		if(!_select[i].active) continue;

		// monitor for exceptions
		if(_select[i].close_cb != NULL) FD_SET(_select[i].fd, &es);

		// store fd to monitor in read and/or write set
		if(_select[i].read_wanted  && _select[i].read_cb  != NULL) FD_SET(_select[i].fd, &rs);
		if(_select[i].write_wanted && _select[i].write_cb != NULL) FD_SET(_select[i].fd, &ws);

		// keep track of the highest fd for the select call
		if(_select[i].fd > max) max = _select[i].fd;
	}

	// wait for a file descriptor to become ready or a timeout
	struct timeval tv = { tv_sec: 0, tv_usec: usec_timeout };
	int n = select(++max, &rs, &ws, &es, usec_timeout<0?NULL:&tv);
	if(n < 0) {
		perror("select");
		return(0);
	}

	// no fd ready
	if(!n) return(1);

	// check which fds are ready to read, write or have an exception
	for(i = 0; i < SELECT_SLOTS_COUNT; i++) {
		// skip inactive slots
		if(!_select[i].active) continue;

		// by default the next section fails
		int failure = 0;

		// check for exception
		if(!FD_ISSET(_select[i].fd, &es)) {
			// on no exception we can read and write at the same cycle
			if(FD_ISSET(_select[i].fd, &rs)) {
				// read on a ready readset
				failure = !_select[i].read_cb(_select[i].ref, _select[i].fd);
			}
			if(FD_ISSET(_select[i].fd, &ws)) {
				// write on a ready writeset
				failure = !_select[i].write_cb(_select[i].ref, _select[i].fd);
			}
		}

		// on failure we just cleanup
		if(failure) {
			// let the client close the connection
			_select[i].close_cb(_select[i].ref, _select[i].fd);

			// cleanup the slot
			_select[i].active = 0;
		}
	}

	return(1);
}

void select_init(void) {
	// clear our memory setting all slots inactive
	memset(&_select, 0, sizeof(_select));
}

